@@ -11,17 +11,49 @@
 # FOR A PARTICULAR PURPOSE.
 #
 ##############################################################################
-"""Security Checkers
+"""
+Security Checkers.
 
-You can set the environment variable ZOPE_WATCH_CHECKERS to get additional
-security checker debugging output on the standard error.
+This module contains the primary implementations of
+:class:`zope.security.interfaces.IChecker` (:class:`Checker`,
+:class:`MultiChecker`, :func:`NamesChecker`) and
+:class:`zope.security.interfaces.IProxyFactory` (:func:`ProxyFactory`).
 
-Setting ZOPE_WATCH_CHECKERS to 1 will display messages about unauthorized or
+It also defines helpers for permission checking (:func:`canAccess`,
+:func:`canWrite`) and getting checkers
+(:func:`getCheckerForInstancesOf`, :func:`selectChecker`).
+
+This module is accelerated with a C implementation on CPython by
+default. If the environment variable ``PURE_PYTHON`` is set (to any
+value) before this module is imported, the C extensions will be
+bypassed and the reference Python implementations will be used. This
+can be helpful for debugging and tracing.
+
+Debugging Permissions Problems
+==============================
+
+You can set the environment variable ``ZOPE_WATCH_CHECKERS`` before
+this module is imported to get additional security checker debugging
+output on the standard error.
+
+Setting ``ZOPE_WATCH_CHECKERS`` to 1 will display messages about unauthorized or
 forbidden attribute access.  Setting it to a larger number will also display
 messages about granted attribute access.
 
-Note that the ZOPE_WATCH_CHECKERS mechanism will eventually be
+Note that the ``ZOPE_WATCH_CHECKERS`` mechanism may eventually be
 replaced with a more general security auditing mechanism.
+
+.. seealso:: :class:`CheckerLoggingMixin`, :class:`WatchingChecker`, :class:`WatchingCombinedChecker`
+
+API
+===
+
+.. py:data:: CheckerPublic
+
+  The special constant that indicates that no permission
+  checking needs to be done.
+
+.. autofunction:: selectChecker
 """
 import abc
 import os
@@ -49,23 +81,23 @@ from zope.security.interfaces import Unauthorized
 from zope.security._definitions import thread_local
 from zope.security._compat import CLASS_TYPES
 from zope.security._compat import PYTHON2
-from zope.security._compat import _u
+from zope.security._compat import PURE_PYTHON
 from zope.security.proxy import Proxy
 from zope.security.proxy import getChecker
 
 try:
     from zope.exceptions import DuplicationError
-except ImportError: #pragma NO COVER
+except ImportError: # pragma: no cover
     class DuplicationError(Exception):
         """A duplicate registration was attempted"""
 
-if os.environ.get('ZOPE_WATCH_CHECKERS'): #pragma NO COVER
+WATCH_CHECKERS = 0
+
+if os.environ.get('ZOPE_WATCH_CHECKERS'):
     try:
         WATCH_CHECKERS = int(os.environ.get('ZOPE_WATCH_CHECKERS'))
-    except ValueError:
+    except ValueError: # pragma: no cover
         WATCH_CHECKERS = 1
-else:
-    WATCH_CHECKERS = 0
 
 
 def ProxyFactory(object, checker=None):
@@ -156,15 +188,22 @@ def canAccess(obj, name):
 
 @implementer(INameBasedChecker)
 class CheckerPy(object):
+    """
+    The Python reference implementation of
+    :class:`zope.security.interfaces.INameBasedChecker`.
+
+    Ordinarily there will be no reason to ever explicitly use this class;
+    instead use the class assigned to :class:`Checker`.
+    """
 
     def __init__(self, get_permissions, set_permissions=None):
         """Create a checker
 
         A dictionary must be provided for computing permissions for
         names. The dictionary get will be called with attribute names
-        and must return a permission id, None, or the special marker,
-        CheckerPublic. If None is returned, then access to the name is
-        forbidden. If CheckerPublic is returned, then access will be
+        and must return a permission ID, None, or the special marker,
+        :const:`CheckerPublic`. If None is returned, then access to the name is
+        forbidden. If :const:`CheckerPublic` is returned, then access will be
         granted without checking a permission.
 
         An optional setattr dictionary may be provided for checking
@@ -256,19 +295,19 @@ class TracebackSupplement(object):
             cls = self.obj.__class__
             if hasattr(cls, "__module__"):
                 s = "%s.%s" % (cls.__module__, cls.__name__)
-            else: #pragma NO COVER XXX
+            else: # pragma: no cover XXX
                 s = str(cls.__name__)
             result.append("   - class: " + s)
-        except: #pragma NO COVER XXX
+        except: # pragma: no cover XXX
             pass
         try:
             cls = type(self.obj)
             if hasattr(cls, "__module__"):
                 s = "%s.%s" % (cls.__module__, cls.__name__)
-            else: #pragma NO COVER XXX
+            else: # pragma: no cover XXX
                 s = str(cls.__name__)
             result.append("   - type: " + s)
-        except: #pragma NO COVER XXX
+        except: # pragma: no cover XXX
             pass
         return "\n".join(result)
 
@@ -282,7 +321,7 @@ class Global(object):
     """
 
     def __init__(self, name, module=None):
-        if module is None: #pragma NO COVER XXX
+        if module is None: # pragma: no cover XXX
             module = sys._getframe(1).f_locals['__name__']
 
         self.__name__ = name
@@ -295,12 +334,14 @@ class Global(object):
         return "%s(%s,%s)" % (self.__class__.__name__,
                               self.__name__, self.__module__)
 
-# Marker for public attributes
+
 CheckerPublic = Global('CheckerPublic')
 CP_HACK_XXX = CheckerPublic
 
 # Now we wrap it in a security proxy so that it retains its
 # identity when it needs to be security proxied.
+# XXX: This means that we can't directly document it with
+# sphinx because issubclass() will fail.
 d = {}
 CheckerPublic = Proxy(CheckerPublic, Checker(d)) # XXX uses CheckerPy
 d['__reduce__'] = CheckerPublic
@@ -316,7 +357,7 @@ def NamesChecker(names=(), permission_id=CheckerPublic, **__kw__):
 
     A sequence of names is given as the first argument. If a second
     argument, permission_id, is given, it is the permission required
-    to access the names.  Additional names and permission ids can be
+    to access the names.  Additional names and permission IDs can be
     supplied as keyword arguments.
     """
 
@@ -330,10 +371,15 @@ def NamesChecker(names=(), permission_id=CheckerPublic, **__kw__):
     return Checker(data)
 
 def InterfaceChecker(interface, permission_id=CheckerPublic, **__kw__):
+    """
+    Create a :func:`NamesChecker` for all the names defined in the *interface*
+    (a subclass of :class:`zope.interface.Interface`).
+    """
     return NamesChecker(interface.names(all=True), permission_id, **__kw__)
 
 def MultiChecker(specs):
-    """Create a checker from a sequence of specifications
+    """
+    Create a checker from a sequence of specifications
 
     A specification is:
 
@@ -341,12 +387,12 @@ def MultiChecker(specs):
 
       o a sequence of names or an interface
 
-      o a permission id
+      o a permission ID
 
       All the names in the sequence of names or the interface are
       protected by the permission.
 
-    - A dictionoid (having an items method), with items that are
+    - A dictionary (having an items method), with items that are
       name/permission-id pairs.
     """
     data = {}
@@ -395,6 +441,7 @@ def selectCheckerPy(object):
 
     while not isinstance(checker, Checker):
         checker = checker(object)
+
         if checker is NoProxy or checker is None:
             return None
 
@@ -408,12 +455,12 @@ DEFINABLE_TYPES = CLASS_TYPES + (types.ModuleType,)
 def defineChecker(type_, checker):
     """Define a checker for a given type of object
 
-    The checker can be a Checker, or a function that, when called with
-    an object, returns a Checker.
+    The checker can be a :class:`Checker`, or a function that, when called with
+    an object, returns a :class:`Checker`.
     """
     if not isinstance(type_, DEFINABLE_TYPES):
         raise TypeError(
-                'type_ must be a type, class or module, not a %s' % type_)
+            'type_ must be a type, class or module, not a %s' % type_)
     if type_ in _checkers:
         raise DuplicationError(type_)
     _checkers[type_] = checker
@@ -439,11 +486,14 @@ _defaultChecker = Checker({})
 _available_by_default = []
 
 # Get optimized versions
-try:
-    import zope.security._zope_security_checker
-except (ImportError, AttributeError): #pragma NO COVER PyPy / PURE_PYTHON
-    pass
-else:
+_c_available = not PURE_PYTHON
+if _c_available:
+    try:
+        import zope.security._zope_security_checker
+    except (ImportError, AttributeError): # pragma: no cover PyPy / PURE_PYTHON
+        _c_available = False
+
+if _c_available:
     from zope.security._zope_security_checker import _checkers, selectChecker
     from zope.security._zope_security_checker import NoProxy, Checker
     from zope.security._zope_security_checker import _defaultChecker
@@ -459,16 +509,23 @@ class CombinedChecker(Checker):
 
     The following table describes the result of a combined checker in detail.
 
-    checker1           checker2           CombinedChecker(checker1, checker2)
-    ------------------ ------------------ -----------------------------------
-    ok                 anything           ok (checker2 is never called)
-    Unauthorized       ok                 ok
-    Unauthorized       Unauthorized       Unauthorized
-    Unauthorized       ForbiddenAttribute Unauthorized
-    ForbiddenAttribute ok                 ok
-    ForbiddenAttribute Unauthorized       Unauthorized
-    ForbiddenAttribute ForbiddenAttribute ForbiddenAttribute
-    ------------------ ------------------ -----------------------------------
+    +--------------------+--------------------+-------------------------------------+
+    | checker1           | checker2           | CombinedChecker(checker1, checker2) |
+    +====================+====================+=====================================+
+    | ok                 | anything           | ok (checker 2 never called)         |
+    +--------------------+--------------------+-------------------------------------+
+    | Unathorized        | ok                 | ok                                  |
+    +--------------------+--------------------+-------------------------------------+
+    | Unauthorized       | Unauthorized       | Unauthorized                        |
+    +--------------------+--------------------+-------------------------------------+
+    | Unauthorized       | ForbiddenAttribute | Unauthorized                        |
+    +--------------------+--------------------+-------------------------------------+
+    | ForbiddenAttribute | ok                 | ok                                  |
+    +--------------------+--------------------+-------------------------------------+
+    | ForbiddenAttribute | Unauthorized       | Unauthorized                        |
+    +--------------------+--------------------+-------------------------------------+
+    | ForbiddenAttribute | ForbiddenAttribute | ForbiddenAttribute                  |
+    +--------------------+--------------------+-------------------------------------+
     """
 
     def __init__(self, checker1, checker2):
@@ -476,6 +533,7 @@ class CombinedChecker(Checker):
         Checker.__init__(self,
                          checker1.get_permissions,
                          checker1.set_permissions)
+
         self._checker2 = checker2
 
     def check(self, object, name):
@@ -498,20 +556,24 @@ class CombinedChecker(Checker):
         except ForbiddenAttribute:
             self._checker2.check_setattr(object, name)
         except Unauthorized as unauthorized_exception:
-            try: self._checker2.check_setattr(object, name)
+            try:
+                self._checker2.check_setattr(object, name)
             except ForbiddenAttribute:
                 raise unauthorized_exception
 
 class CheckerLoggingMixin(object):
-    """Debugging mixin for checkers.
+    """
+    Debugging mixin for checkers.
 
     Prints verbose debugging information about every performed check to
-    sys.stderr.
+    :data:`sys.stderr`.
 
-    If verbosity is set to 1, only displays Unauthorized and Forbidden messages.
-    If verbosity is set to a larger number, displays all messages.
     """
 
+    #: If set to 1 (the default), only displays ``Unauthorized`` and
+    #: ``Forbidden`` messages. If verbosity is set to a larger number,
+    #: displays all messages. Normally this is controlled via the environment
+    #: variable ``ZOPE_WATCH_CHECKERS``.
     verbosity = 1
     _file = sys.stderr
 
@@ -525,7 +587,7 @@ class CheckerLoggingMixin(object):
             if self.verbosity > 1:
                 if name in _available_by_default:
                     self._log('[CHK] + Always available: %s on %r'
-                        % (name, object), 2)
+                              % (name, object), 2)
                 else:
                     self._log(
                         '[CHK] + Granted: %s on %r' % (name, object), 2)
@@ -577,54 +639,90 @@ class CheckerLoggingMixin(object):
             raise
 
 
-if WATCH_CHECKERS: #pragma NO COVER
-    class Checker(CheckerLoggingMixin, Checker):
-        verbosity = WATCH_CHECKERS
-    class CombinedChecker(CheckerLoggingMixin, CombinedChecker):
-        verbosity = WATCH_CHECKERS
+# We have to be careful with the order of inheritance
+# here. See https://github.com/zopefoundation/zope.security/issues/8
+class WatchingChecker(CheckerLoggingMixin, Checker):
+    """
+    A checker that will perform verbose logging. This will be set
+    as the default when ``ZOPE_WATCH_CHECKERS`` is set when this
+    module is imported.
+    """
+    verbosity = WATCH_CHECKERS
+class WatchingCombinedChecker(CombinedChecker, WatchingChecker):
+    """
+    A checker that will perform verbose logging. This will be set
+    as the default when ``ZOPE_WATCH_CHECKERS`` is set when this
+    module is imported.
+    """
+    verbosity = WATCH_CHECKERS
+
+if WATCH_CHECKERS: # pragma: no cover
+    # When we make these the default, we also need to be sure
+    # to update the _defaultChecker's type (if it's not the C
+    # extension) so that selectCheckerPy can properly recognize
+    # it as a Checker.
+    # See https://github.com/zopefoundation/zope.security/issues/8
+    Checker = WatchingChecker
+    CombinedChecker = WatchingCombinedChecker
+
+    if not _c_available:
+        _defaultChecker.__class__ = Checker
 
 def _instanceChecker(inst):
     return _checkers.get(inst.__class__, _defaultChecker)
 
 def moduleChecker(module):
+    """
+    Return the :class:`zope.security.interfaces.IChecker` defined for the
+    *module*, if any.
+
+    .. seealso:: :func:`zope.security.metaconfigure.protectModule`
+       To define module protections.
+    """
     return _checkers.get(module)
 
 
-_available_by_default[:] = ['__lt__', '__le__', '__eq__',
-                            '__gt__', '__ge__', '__ne__',
-                            '__hash__', '__nonzero__',
-                            '__class__', '__providedBy__', '__implements__',
-                            '__repr__', '__conform__',
-                            '__name__', '__parent__',
-                            ]
+_available_by_default[:] = [
+    '__lt__', '__le__', '__eq__',
+    '__gt__', '__ge__', '__ne__',
+    '__hash__', '__nonzero__',
+    '__class__', '__providedBy__', '__implements__',
+    '__repr__', '__conform__',
+    '__name__', '__parent__',
+]
 
 _callableChecker = NamesChecker(['__str__', '__name__', '__call__'])
-_typeChecker = NamesChecker(
-    ['__str__', '__name__', '__module__', '__bases__', '__mro__',
-     '__implemented__'])
+_typeChecker = NamesChecker([
+    '__str__', '__name__', '__module__', '__bases__', '__mro__',
+    '__implemented__',
+])
 _namedChecker = NamesChecker(['__name__'])
+_iteratorChecker = NamesChecker([
+    'next', '__next__', '__iter__', '__len__',
+    '__length_hint__',
+])
 
-_iteratorChecker = NamesChecker(['next', '__next__', '__iter__', '__len__'])
+_setChecker = NamesChecker([
+    '__iter__', '__len__', '__str__', '__contains__',
+    'copy', 'difference', 'intersection', 'issubset',
+    'issuperset', 'symmetric_difference', 'union',
+    '__and__', '__or__', '__sub__', '__xor__',
+    '__rand__', '__ror__', '__rsub__', '__rxor__',
+    '__eq__', '__ne__', '__lt__', '__gt__',
+    '__le__', '__ge__',
+])
 
-_setChecker = NamesChecker(['__iter__', '__len__', '__str__', '__contains__',
-                            'copy', 'difference', 'intersection', 'issubset',
-                            'issuperset', 'symmetric_difference', 'union',
-                            '__and__', '__or__', '__sub__', '__xor__',
-                            '__rand__', '__ror__', '__rsub__', '__rxor__',
-                            '__eq__', '__ne__', '__lt__', '__gt__',
-                            '__le__', '__ge__'])
-
-class BasicTypes(dict):
+class _BasicTypes(dict):
     """Basic Types Dictionary
 
     Make sure that checkers are really updated, when a new type is added.
     """
     def __setitem__(self, name, value):
-        super(BasicTypes.__class__, self).__setitem__(name, value)
+        dict.__setitem__(self, name, value)
         _checkers[name] = value
 
     def __delitem__(self, name):
-        super(BasicTypes.__class__, self).__delitem__(name)
+        dict.__delitem__(self, name)
         del _checkers[name]
 
     def clear(self):
@@ -632,7 +730,7 @@ class BasicTypes(dict):
         raise NotImplementedError
 
     def update(self, d):
-        super(BasicTypes.__class__, self).update(d)
+        dict.update(self, d)
         _checkers.update(d)
 
 _basic_types = {
@@ -654,19 +752,19 @@ _basic_types = {
 if PYTHON2:
     _basic_types[long] = NoProxy
     _basic_types[unicode] = NoProxy
-else: #pragma NO COVER
+else: # pragma: no cover
     _basic_types[type({}.values())] = NoProxy
     _basic_types[type({}.keys())] = NoProxy
     _basic_types[type({}.items())] = NoProxy
 
 try:
     import pytz
-except ImportError:
+except ImportError:  # pragma: no cover
     pass
-else: #pragma NO COVER
+else:
     _basic_types[type(pytz.UTC)] = NoProxy
 
-BasicTypes = BasicTypes(_basic_types)
+BasicTypes = _BasicTypes(_basic_types)
 del _basic_types
 
 # Available for tests. Located here so it can be kept in sync with BasicTypes.
@@ -686,13 +784,15 @@ BasicTypes_examples = {
 }
 
 if PYTHON2:
-    BasicTypes_examples[unicode] = _u('uabc')
+    BasicTypes_examples[unicode] = u'uabc'
     BasicTypes_examples[long] = long(65536)
 
 
-class _Sequence(object): #pragma NO COVER
-    def __len__(self): return 0
-    def __getitem__(self, i): raise IndexError
+class _Sequence(object):
+    def __len__(self):
+        raise NotImplementedError()
+    def __getitem__(self, i):
+        raise NotImplementedError()
 
 _Declaration_checker = InterfaceChecker(
     IDeclaration,
@@ -702,7 +802,7 @@ _Declaration_checker = InterfaceChecker(
     __call__=CheckerPublic,
     )
 
-def f(): #pragma NO COVER
+def f(): # pragma: no cover
     yield f
 
 
@@ -716,20 +816,27 @@ _default_checkers = {
                         '__add__', '__radd__', ]),
     set: _setChecker,
     frozenset: _setChecker,
+    # XXX: actually decimal.Decimal has more methods, which are unlisted here
+    #      so expect ForbiddenAttribute on such
     decimal.Decimal: NamesChecker(['__nonzero__', '__cmp__', '__eq__',
-                                   '__ne__', 'compare', '__hash__',
-                                   'as_tuple', '__str__', 'to_eng_string',
+                                   '__ne__', '__hash__',
+                                   '__str__',
                                    '__neg__', '__pos__', '__abs__',
-                                   '__add__', '__radd__', '__sub__',
-                                   '__rsub__', '__mul__', '__rmul__',
-                                   '__div__', '__rdiv__', '__rtruediv__',
-                                   '__divmod__', '__rdivmod__', '__mod__',
-                                   '__rmod__', 'remainder_near',
+                                   '__add__', '__radd__',
+                                   '__sub__', '__rsub__',
+                                   '__mul__', '__rmul__',
+                                   '__div__', '__truediv__',
+                                   '__rdiv__', '__rtruediv__',
+                                   '__divmod__', '__rdivmod__',
+                                   '__mod__', '__rmod__',
                                    '__floordiv__', '__rfloordiv__',
                                    '__float__', '__int__', '__long__',
-                                   '__pow__', '__rpow__', 'normalize',
-                                   'quantize', 'same_quantum', 'to_integral',
-                                   'sqrt', 'max', 'min', 'adjusted']),
+                                   '__pow__', '__rpow__',
+                                   'adjusted', 'as_tuple', 'compare',
+                                   'max', 'min', 'normalize',
+                                   'quantize', 'remainder_near',
+                                   'same_quantum', 'sqrt',
+                                   'to_eng_string', 'to_integral' ]),
 
     # YAGNI: () a rock
     tuple: NamesChecker(['__getitem__', '__getslice__', '__add__', '__radd__',
@@ -754,7 +861,7 @@ _default_checkers = {
         __str__=CheckerPublic, _implied=CheckerPublic, subscribe=CheckerPublic,
         ),
     zope.interface.interface.Method: InterfaceChecker(
-                                        zope.interface.interfaces.IMethod),
+        zope.interface.interfaces.IMethod),
     zope.interface.declarations.ProvidesClass: _Declaration_checker,
     zope.interface.declarations.ClassProvides: _Declaration_checker,
     zope.interface.declarations.Implements: _Declaration_checker,
@@ -779,6 +886,10 @@ def _fixup_dictlike(dict_type):
             o_type = type(obj)
             if o_type not in _default_checkers:
                 _default_checkers[o_type] = _iteratorChecker
+            # PyPy3 has special types for iter({}.items()) etc.
+            iter_type = type(iter(obj))
+            if iter_type not in _default_checkers:
+                _default_checkers[iter_type] = _iteratorChecker
 
 def _fixup_odict():
     # OrderedDicts have three different implementations: Python 2 (pure
@@ -823,6 +934,111 @@ else:
 
 del _fixup_dictlike
 
+def _fixup_zope_interface():
+    # Make sure the provided and implementedBy objects
+    # can be iterated.
+    # Note that we DO NOT use the _iteratorChecker, but instead
+    # we use NoProxy to be sure that the results (of iteration or not) are not
+    # proxied. On Python 2, these objects are builtin and don't go through the
+    # checking process at all, much like BTrees, so NoProxy is necessary for
+    # compatibility. On Python 3, prior to this, iteration was simply not allowed.
+    from zope.interface import providedBy
+    from zope.interface import alsoProvides
+
+    class I1(Interface):
+        pass
+
+    class I2(Interface):
+        pass
+
+    @implementer(I1)
+    class O(object):
+        pass
+
+    o = O()
+
+
+    # This will be athe zope.interface.implementedBy from the class
+    # a zope.interface.declarations.Implements
+    _default_checkers[type(providedBy(o))] = NoProxy
+
+    alsoProvides(o, I2)
+    # This will be the zope.interface.Provides from the instance
+    _default_checkers[type(providedBy(o))] = NoProxy
+
+_fixup_zope_interface()
+del _fixup_zope_interface
+
+
+def _fixup_itertools():
+    # itertools.groupby is a built-in custom iterator type introduced
+    # in python2.4. It should have the same checker as other built-in
+    # iterators.
+
+    # Also, itertools._grouper also needs to be exposed as an
+    # iterator. Its type is not exposed by name, but can be accessed
+    # like so: type(list(itertools.groupby([0]))[0][1])
+
+    import itertools
+
+    group = itertools.groupby([0])
+    type_group = type(group)
+    if type_group not in _default_checkers:
+        _default_checkers[type_group] = _iteratorChecker
+
+    type_grouper = type(list(group)[0][1])
+    if type_grouper not in _default_checkers:
+        _default_checkers[type_grouper] = _iteratorChecker
+
+    # There are also many other custom types in itertools that need the
+    # same treatment. See a similar list in test_checker.py:test_itertools_checkers
+    pred = lambda x: x
+    iterable = (1, 2, 3)
+    pred_iterable = (pred, iterable)
+    missing_in_py3 = {'ifilter', 'ifilterfalse', 'imap',
+                      'izip', 'izip_longest'}
+    missing_in_py2 = {'zip_longest', 'accumulate', 'compress',
+                      'combinations', 'combinations_with_replacement'}
+    missing = missing_in_py3 if sys.version_info[0] >= 3 else missing_in_py2
+    for func, args in (
+            ('count', ()),
+            ('cycle', ((),)),
+            ('dropwhile', pred_iterable),
+            ('ifilter', pred_iterable),
+            ('ifilterfalse', pred_iterable),
+            ('imap', pred_iterable),
+            ('islice', (iterable, 2)),
+            ('izip', (iterable,)),
+            ('izip_longest', (iterable,)),
+            ('permutations', (iterable,)),
+            ('product', (iterable,)),
+            ('repeat', (1, 2)),
+            ('starmap', pred_iterable),
+            ('takewhile', pred_iterable),
+            ('tee', (iterable,)),
+            # Python 3 additions
+            ('zip_longest', (iterable,)),
+            ('accumulate', (iterable,)),
+            ('compress', (iterable, ())),
+            ('combinations', (iterable, 1)),
+            ('combinations_with_replacement', (iterable, 1)),
+    ):
+        try:
+            func = getattr(itertools, func)
+        except AttributeError:
+            assert func in missing, "Expected %s but not found" % (func,)
+            continue # pragma: no cover This is hit on Python 2, but it doesn't always show
+
+        result = func(*args)
+        if func == itertools.tee:
+            result = result[0]
+        tresult = type(result)
+        if tresult not in _default_checkers:
+            _default_checkers[tresult] = _iteratorChecker
+
+_fixup_itertools()
+del _fixup_itertools
+
 def _clear():
     _checkers.clear()
     _checkers.update(_default_checkers)
@@ -832,7 +1048,7 @@ _clear()
 
 try:
     from zope.testing.cleanup import addCleanUp
-except ImportError: #pragma NO COVER
+except ImportError: # pragma: no cover
     pass
 else:
     addCleanUp(_clear)
